/*############################################################################
  # Copyright (C) 2005 Intel Corporation
  #
  # SPDX-License-Identifier: MIT
  ############################################################################*/

#define PRODUCT_NAME           "Intel\xae Media SDK"
#define FILE_VERSION           1, 0, 0, 0
#define FILE_VERSION_STRING    "1,0,0,0"
#define FILTER_NAME_PREFIX     ""
#define FILTER_NAME_SUFFIX     ""
#define PRODUCT_COPYRIGHT      "Copyright\xa9 2003-2018 Intel Corporation"
#define PRODUCT_VERSION        1, 0, 0, 0
#define PRODUCT_VERSION_STRING "1,0,0,0"
