/*############################################################################
  # Copyright (C) 2005 Intel Corporation
  #
  # SPDX-License-Identifier: MIT
  ############################################################################*/

#ifndef __MFX_SAMPLES_CONFIG__
#define __MFX_SAMPLES_CONFIG__

#if defined(_WIN32) || defined(_WIN64)
    #pragma warning(disable : 4068) // unknown pragma
    #pragma warning(disable : 4996) // declared deprecated
    #pragma warning(disable : 4163) // 'identifier' : not available as an intrinsic function
#endif

#endif //__MFX_SAMPLES_CONFIG__
